import numpy as np
from tensorflow.python.framework import ops
from tensorflow.python.framework import graph_util
from tensorflow.python.profiler.internal.flops_registry import _reduction_op_flops


@ops.RegisterStatistics("FusedBatchNormV3", "flops")
def _flops_fused_batch_norm_v3(graph, node):
    """inference is only supportted"""
    in_shape = graph_util.tensor_shape_from_node_def_name(graph, node.input[0])
    in_shape.assert_is_fully_defined()
    mean_shape = graph_util.tensor_shape_from_node_def_name(graph, node.input[3])
    mean_shape.assert_is_fully_defined()
    variance_shape = graph_util.tensor_shape_from_node_def_name(graph, node.input[4])
    variance_shape.assert_is_fully_defined()

    if node.attr["is_training"].b is True:
        raise ValueError("Only supports inference mode")

    num_flops = (
        in_shape.num_elements()
        + 4 * variance_shape.num_elements()
        + mean_shape.num_elements()
    )
    return ops.OpStats("flops", num_flops)


@ops.RegisterStatistics("Max", "flops")
def _flops_max(graph, node):
    """inference is supportted"""
    # reduction - comparison, no finalization
    return _reduction_op_flops(graph, node, reduce_flops=1, finalize_flops=0)




from typing import Optional, Union
import tensorflow as tf
from tensorflow.python.framework.convert_to_constants import (
    convert_variables_to_constants_v2_as_graph,
)

from tensorflow.keras import Sequential, Model

import keras_flops.flops_registory


def get_flops(model: Union[Model, Sequential], batch_size: Optional[int] = None) -> int:
    """
    Calculate FLOPS for tf.keras.Model or tf.keras.Sequential .
    Ignore operations used in only training mode such as Initialization.
    Use tf.profiler of tensorflow v1 api.
    """
    if not isinstance(model, (Sequential, Model)):
        raise KeyError(
            "model arguments must be tf.keras.Model or tf.keras.Sequential instanse"
        )

    if batch_size is None:
        batch_size = 1

    # convert tf.keras model into frozen graph to count FLOPS about operations used at inference
    # FLOPS depends on batch size
    inputs = [
        tf.TensorSpec([batch_size] + inp.shape[1:], inp.dtype) for inp in model.inputs
    ]
    real_model = tf.function(model).get_concrete_function(inputs)
    frozen_func, _ = convert_variables_to_constants_v2_as_graph(real_model)

    # Calculate FLOPS with tf.profiler
    run_meta = tf.compat.v1.RunMetadata()
    opts = tf.compat.v1.profiler.ProfileOptionBuilder.float_operation()
    flops = tf.compat.v1.profiler.profile(
        graph=frozen_func.graph, run_meta=run_meta, cmd="scope", options=opts
    )
    # print(frozen_func.graph.get_operations())
    # TODO: show each FLOPS
    return flops.total_float_ops



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['keras_flops']

package_data = \
{'': ['*']}

install_requires = \
['tensorflow>=2.2,<3.0']

setup_kwargs = {
    'name': 'keras-flops',
    'version': '0.1.0',
    'description': 'FLOPs calculator for neural network architecture written in tensorflow 2.x (tf.keras)',
    'long_description': '# keras-flops\n\n![](https://github.com/tokusumi/keras-flops/workflows/Tests/badge.svg)\n\nFLOPs calculator for neural network architecture written in tensorflow (tf.keras) v2.2+\n\nThis stands on the shoulders of giants, [tf.profiler](https://www.tensorflow.org/api_docs/python/tf/compat/v1/profiler/Profiler). \n\n## Requirements\n\n* Python 3.6+\n* Tensorflow 2.2+\n\n## Installation\n\nThis implementation is simple thanks to stands on the shoulders of giants, [tf.profiler](https://www.tensorflow.org/api_docs/python/tf/compat/v1/profiler/Profiler). Only one function is defined.\n\nCopy and paste [it](https://github.com/tokusumi/keras-flops/blob/master/keras_flops/flops_calculation.py).\n\n## Example\n\nSee colab examples [here](https://github.com/tokusumi/keras-flops/tree/master/notebooks) in details.\n\n```python\nfrom tensorflow.keras import Model, Input\nfrom tensorflow.keras.layers import Dense, Flatten, Conv2D, MaxPooling2D, Dropout\n\nfrom keras_flops import get_flops\n\n# build model\ninp = Input((32, 32, 3))\nx = Conv2D(32, kernel_size=(3, 3), activation="relu")(inp)\nx = Conv2D(64, (3, 3), activation="relu")(x)\nx = MaxPooling2D(pool_size=(2, 2))(x)\nx = Dropout(0.25)(x)\nx = Flatten()(x)\nx = Dense(128, activation="relu")(x)\nx = Dropout(0.5)(x)\nout = Dense(10, activation="softmax")(x)\nmodel = Model(inp, out)\n\n# Calculae FLOPS\nflops = get_flops(model, batch_size=1)\nprint(f"FLOPS: {flops / 10 ** 9:.03} G")\n# >>> FLOPS: 0.0338 G\n```\n\n## Support\n\nSupport `tf.keras.layers` as follows,\n\n| name | layer | \n| -- | -- |\n| Conv | Conv[1D/2D/3D]|\n| | Conv[1D/2D]Transpose |\n| | DepthwiseConv2D |\n| | SeparableConv[1D/2D] |\n| Pooling | AveragePooling[1D/2D] |\n| | GlobalAveragePooling[1D/2D/3D]|\n| | MaxPooling[1D/2D] |\n| | GlobalMaxPool[1D/2D/3D] |\n| Normalization | BatchNormalization |\n| Activation | Softmax |\n| Attention | Attention |\n| | AdditiveAttention |\n| others | Dense |\n\n## Not supported\n\nNot support `tf.keras.layers` as follows. They are calculated as zero or smaller value than correct value.\n\n| name | layer | \n| -- | -- |\n| Conv | Conv3DTranspose |\n| Pooling | AveragePooling3D |\n| | MaxPooling3D |\n| | UpSampling[1D/2D/3D] |\n| Normalization | LayerNormalization |\n| RNN | SimpleRNN |\n| | LSTM |\n| | GRU |\n| others | Embedding |',
    'author': 'tokusumi',
    'author_email': 'tksmtoms@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tokusumi/keras-flops',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
